import pytest

from galacteek.core import jsonf

class File1(jsonf.QJSONFile):
    def prepare(self, root):
        root['comments'] = {
            '1': 'This is it',
        }

@pytest.fixture
def jfile1(tmpdir):
    return File1(tmpdir.join('jsonf1.json'))

class TestMarks:
    def test_simple(self, jfile1):
        jfile1.save()
        print(jfile1.root)




import sys
import json

import aioipfs
import asyncio

from galacteek.ipfs.pubsubmsg import MarksBroadcastMessage

class PubsubListener(object):
    """ IPFS pubsub listener for a given topic """

    def __init__(self, client, loop, ipfsCtx, topic='defaulttopic'):
        self.ipfsCtx = ipfsCtx
        self.loop = loop
        self.topic = topic
        self.client = client
        self.inqueue = asyncio.Queue()
        self.lock = asyncio.Lock()

        self.tsServe = None
        self.tsProcess = None
        self.tsPeriodic = None

    def msgDataJson(self, msg):
        """
        Decode JSON data contained in a pubsub message
        """
        try:
            return json.loads(msg['data'].decode())
        except Exception as e:
            print('Could not decode {}'.format(
                msg['data'], file=sys.stderr))
            return None

    def start(self):
        self.tsServe = self.loop.create_task(self.serve(self.client))
        self.tsProcess = self.loop.create_task(self.processMessages())
        self.tsPeriodic = self.loop.create_task(self.periodic())

    def stop(self):
        for ts in [ self.tsServe, self.tsProcess, self.tsPeriodic ]:
            if ts: ts.cancel()

    async def processMessages(self):
        return True

    async def periodic(self):
        return True

    async def serve(self, client):
        nodeInfo = await client.core.id()
        nodeId = nodeInfo['ID']

        async with client as cli:
            async for message in cli.pubsub.sub(self.topic):
                await asyncio.sleep(0)
                if message['from'] == nodeId:
                    continue

                self.ipfsCtx.pubsubMessageRx.emit()
                await self.inqueue.put(message)

    async def send(self, data, topic=None):
        if topic is None:
            topic = self.topic
        status = await self.client.pubsub.pub(topic, data)
        self.ipfsCtx.pubsubMessageTx.emit()
        return status

class HashmarksExchanger(PubsubListener):
    def __init__(self, client, loop, ipfsCtx, marksLocal, marksNetwork):
        super().__init__(client, loop, ipfsCtx, topic='galacteek.ipfsmarks')
        self.marksLocal = marksLocal
        self.marksNetwork = marksNetwork

    async def periodic(self):
        """ Very basic broadcasting for now """
        while True:
            await asyncio.sleep(60)
            o = self.marksLocal.getAll(share=True)
            if len(o.keys()) == 0:
                continue
            msg = MarksBroadcastMessage.make(o)
            await self.send(str(msg))

    async def processMessages(self):
        while True:
            data = await self.inqueue.get()

            msg = self.msgDataJson(data)
            if not msg:
                continue

            msgType = msg.get('msgtype', None)
            if msgType == MarksBroadcastMessage.TYPE:
                await self.processBroadcast(msg)

            await asyncio.sleep(0)

    async def processBroadcast(self, msg):
        marks = msg.get('marks', None)
        if not marks:
            return

        addedCount = 0
        for mark in marks.items():
            await asyncio.sleep(0)
            mPath = mark[0]
            if self.marksNetwork.search(mPath):
                continue
            self.marksNetwork.insertMark(mark, 'auto')
            addedCount += 1

        if addedCount > 0:
            self.ipfsCtx.pubsubMarksReceived.emit(addedCount)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'galacteek/ui/donatedialog.ui'
#
# Created by: PyQt5 UI code generator 5.10
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_DonateDialog(object):
    def setupUi(self, DonateDialog):
        DonateDialog.setObjectName("DonateDialog")
        DonateDialog.resize(875, 293)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(DonateDialog)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.textBrowser = QtWidgets.QTextBrowser(DonateDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.textBrowser.sizePolicy().hasHeightForWidth())
        self.textBrowser.setSizePolicy(sizePolicy)
        self.textBrowser.setMaximumSize(QtCore.QSize(16777215, 200))
        self.textBrowser.setObjectName("textBrowser")
        self.verticalLayout.addWidget(self.textBrowser)
        self.label = QtWidgets.QLabel(DonateDialog)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.moneroAddress = QtWidgets.QLabel(DonateDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.moneroAddress.setFont(font)
        self.moneroAddress.setText("")
        self.moneroAddress.setTextFormat(QtCore.Qt.RichText)
        self.moneroAddress.setWordWrap(False)
        self.moneroAddress.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByMouse)
        self.moneroAddress.setObjectName("moneroAddress")
        self.horizontalLayout.addWidget(self.moneroAddress)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.moneroClip = QtWidgets.QPushButton(DonateDialog)
        self.moneroClip.setText("")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/share/icons/clipboard.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.moneroClip.setIcon(icon)
        self.moneroClip.setObjectName("moneroClip")
        self.horizontalLayout.addWidget(self.moneroClip)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.verticalLayout_2.addLayout(self.verticalLayout)
        self.label_3 = QtWidgets.QLabel(DonateDialog)
        self.label_3.setObjectName("label_3")
        self.verticalLayout_2.addWidget(self.label_3)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.bitcoinAddress = QtWidgets.QLabel(DonateDialog)
        self.bitcoinAddress.setText("")
        self.bitcoinAddress.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByMouse)
        self.bitcoinAddress.setObjectName("bitcoinAddress")
        self.horizontalLayout_2.addWidget(self.bitcoinAddress)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem1)
        self.bitcoinClip = QtWidgets.QPushButton(DonateDialog)
        self.bitcoinClip.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.bitcoinClip.setText("")
        self.bitcoinClip.setIcon(icon)
        self.bitcoinClip.setObjectName("bitcoinClip")
        self.horizontalLayout_2.addWidget(self.bitcoinClip)
        self.verticalLayout_2.addLayout(self.horizontalLayout_2)
        self.okButton = QtWidgets.QPushButton(DonateDialog)
        self.okButton.setMinimumSize(QtCore.QSize(200, 0))
        self.okButton.setObjectName("okButton")
        self.verticalLayout_2.addWidget(self.okButton, 0, QtCore.Qt.AlignRight)

        self.retranslateUi(DonateDialog)
        QtCore.QMetaObject.connectSlotsByName(DonateDialog)

    def retranslateUi(self, DonateDialog):
        _translate = QtCore.QCoreApplication.translate
        DonateDialog.setWindowTitle(_translate("DonateDialog", "Dialog"))
        self.textBrowser.setHtml(_translate("DonateDialog", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'Sans Serif\'; font-size:9pt; font-weight:400; font-style:normal;\">\n"
"<p align=\"center\" style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:10pt;\">Any donation is most welcome to help keep the project\'s development going. The application\'s translation to other languages is planned as well as new features. Donations will help support other projects like </span><a href=\"https://ipfs-search.com\"><span style=\" text-decoration: underline; color:#0000ff;\">IPFS Search</span></a></p>\n"
"<p align=\"center\" style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-size:10pt;\"><br /></p>\n"
"<p align=\"center\" style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:10pt;\">You can make a donation with Monero (or bitcoin) using the addresses below.</span></p>\n"
"<p align=\"center\" style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-size:10pt;\"><br /></p>\n"
"<p align=\"center\" style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:10pt;\">Thanks!</span></p></body></html>"))
        self.label.setText(_translate("DonateDialog", "Monero donation address"))
        self.label_3.setText(_translate("DonateDialog", "Bitcoin donation address"))
        self.okButton.setText(_translate("DonateDialog", "OK"))

from . import galacteek_rc




import asyncio

import aioipfs

class Pinner(object):
    """ Pins objects on request through an async queue """

    def __init__(self, app):
        self.pintasks = []
        self.app = app
        self.queue = asyncio.Queue(loop=self.app.loop)
        self.lock = asyncio.Lock()
        self.pinstatus = {}

    def status(self):
        return self.pinstatus

    def _emitSLength(self):
        self.app.ipfsCtx.pinItemsCount.emit(len(self.pinstatus))

    def pathRegister(self, path):
        self.pinstatus[path] = {}
        self.app.ipfsCtx.pinNewItem.emit(path)
        self._emitSLength()

    def pathDelete(self, path):
        if path in self.pinstatus:
            del self.pinstatus[path]
        self._emitSLength()

    async def pin(self, path, recursive=False):
        ipfsClient = self.app.ipfsClient
        self.pathRegister(path)

        async for pinned in ipfsClient.pin.add(path, recursive=recursive):
            await asyncio.sleep(0)
            async with self.lock:
                self.pinstatus[path] = pinned
            self.app.ipfsCtx.pinItemStatusChanged.emit(path, pinned)

        async with self.lock:
            self.pathDelete(path)

        self.app.ipfsCtx.pinFinished.emit(path)
        return path

    async def enqueue(self, path, recursive, onSuccess):
        await self.queue.put((path, recursive, onSuccess))
        self.app.ipfsCtx.pinQueueSizeChanged.emit(self.queue.qsize())

    async def process(self):
        while True:
            item = await self.queue.get()
            if item is None:
                continue
            self.app.ipfsCtx.pinQueueSizeChanged.emit(self.queue.qsize())
            path, recursive, callback = item
            f = asyncio.ensure_future(self.pin(path, recursive=recursive))
            if callback:
                f.add_done_callback(callback)



# Galacteek, GUI startup entry point

import asyncio
import sys
import argparse
import shutil

from PyQt5.QtWidgets import QApplication

from galacteek.ipfs import ipfsd
from galacteek.ui import mainui
from galacteek.ui.i18n import *
from galacteek import application
from galacteek.appsettings import *

def galacteekGui(args):
    gApp = application.GalacteekApplication(profile=args.profile,
            debug=args.debug)
    loop = gApp.setupAsyncLoop()
    sManager = gApp.settingsMgr

    section = CFG_SECTION_IPFSD
    if args.apiport:
        sManager.setSetting(section, CFG_KEY_APIPORT, args.apiport)
    if args.swarmport:
        sManager.setSetting(section, CFG_KEY_SWARMPORT, args.swarmport)
    if args.gatewayport:
        sManager.setSetting(section, CFG_KEY_HTTPGWPORT, args.gatewayport)

    if sManager.isTrue(CFG_SECTION_IPFSD, CFG_KEY_ENABLED):
        fsMigratePath = shutil.which('fs-repo-migrations')
        hasFsMigrate = fsMigratePath is not None

        if hasFsMigrate is False and args.migrate is True:
            gApp.systemTrayMessage('Galacteek', iFsRepoMigrateNotFound())

        enableMigrate = hasFsMigrate is True and args.migrate is True

        # Look if we can find the ipfs executable
        ipfsPath = shutil.which('ipfs')
        if not ipfsPath:
            gApp.systemTrayMessage('Galacteek', iGoIpfsNotFound(), timeout=8000)
        else:
            gApp.startIpfsDaemon(migrateRepo=enableMigrate)
    else:
        gApp.updateIpfsClient()

    gApp.startPinner()
    gApp.mainWindow.addHashmarksTab()

    # Use the context manager so loop cleanup/close is automatic
    with loop:
        loop.run_forever()

def start():
    parser = argparse.ArgumentParser()

    parser.add_argument('--apiport',  default=None,
        help='IPFS API port number')
    parser.add_argument('--swarmport', default=None,
        help='IPFS swarm port number')
    parser.add_argument('--gatewayport', default=None,
        help='IPFS http gateway port number')
    parser.add_argument('--profile', default='main',
        help='Application Profile')
    parser.add_argument('--migrate', action='store_true',
        dest='migrate', help = 'Activate automatic repository migration')
    parser.add_argument('-d', action='store_true',
        dest='debug', help = 'Activate debugging')
    args = parser.parse_args()

    galacteekGui(args)




import multiprocessing
import subprocess
import os, os.path
import json
import signal
import asyncio

from yarl import URL

def ipfs_config(param, value):
    os.system("ipfs config '{0}' '{1}'".format(param, value))

def ipfs_config_json(param, value):
    os.system('ipfs config --json "{0}" "{1}"'.format(
        param, json.dumps(value)))

DEFAULT_APIPORT = 5001
DEFAULT_SWARMPORT = 4001
DEFAULT_GWPORT = 8080

class IPFSMultiDaemon(multiprocessing.Process):
    """ Runs an IPFS daemon in a separate multiprocessing.Process """

    def __init__(self, repopath, apiport=DEFAULT_APIPORT,
            swarmport=DEFAULT_SWARMPORT,
            gatewayport=DEFAULT_GWPORT, initrepo=True,
            swarm_lowwater=30, swarm_highwater=50,
            queue=None):

        if not queue:
            queue = multiprocessing.Queue()

        self.repopath = repopath
        self.apiport = apiport
        self.gatewayport = gatewayport
        self.swarmport = swarmport
        self.swarm_lowwater = swarm_lowwater
        self.swarm_highwater = swarm_highwater
        self.initrepo = initrepo

        self.proc_pid = None
        self.queue = queue

        multiprocessing.Process.__init__(self)

    def run(self):
        # Sets the IPFS_PATH environment variable
        os.putenv('IPFS_PATH', self.repopath)
        if not os.path.isdir(self.repopath):
            os.mkdir(self.repopath)

        if not os.path.exists(os.path.join(self.repopath, 'config')):
            os.system('ipfs init')

        # Change the ports we listen on
        ipfs_config('Addresses.API',
                '/ip4/127.0.0.1/tcp/{0}'.format(self.apiport))
        ipfs_config('Addresses.Gateway',
                '/ip4/127.0.0.1/tcp/{0}'.format(self.gatewayport))
        ipfs_config_json('Addresses.Swarm',
                '["/ip4/0.0.0.0/tcp/{0}"]'.format(self.swarmport))

        # Swarm connection manager parameters
        ipfs_config_json('Swarm.ConnMgr.LowWater', self.swarm_lowwater)
        ipfs_config_json('Swarm.ConnMgr.HighWater', self.swarm_highwater)

        # Runs the go-ipfs daemon with pubsub activated
        self.sp = subprocess.Popen(['ipfs', 'daemon',
            '--enable-pubsub-experiment'],
             stdout=subprocess.PIPE,
             stderr=subprocess.PIPE)

        self.proc_pid = self.sp.pid
        self.queue.put(('Running', self.proc_pid))
        #(stdout, stderr)  = self.sp.communicate()

        for j in self.queue.get():
            if j == 'S':
                self.sp.terminate()
                self.queue.put('STOPPED')

    def getGatewayPort(self):
        return self.gatewayport

    def getGatewayUrl(self):
        return URL.build(
                host='localhost',
                port=self.getGatewayPort(),
                scheme='http',
                path='')

    def stop(self):
        self.queue.put('S')
        stopped = self.queue.get()

    def killproc(self):
        if self.proc_pid:
            try:
                os.kill(self.proc_pid, signal.SIGINT)
            except Exception as e:
                print('Could not stop the ipfs daemon', file=sys.stderr)




import pytest

import tempfile
import time
import os
import json

import asyncio
import aioipfs

from galacteek.ipfs import asyncipfsd
from galacteek.ipfs import ipfsdconfig
from galacteek.ipfs.ipfsops import *
from galacteek.ipfs.tunnel import *

apiport = 9005
gwport = 9081
swarmport = 9003

@pytest.fixture()
def ipfsdaemon(event_loop, tmpdir):
    dir = tmpdir.mkdir('ipfsdaemon')
    daemon = asyncipfsd.AsyncIPFSDaemon(dir,
            apiport=apiport,
            gatewayport=gwport,
            swarmport=swarmport,
            loop=event_loop,
            pubsubEnable=False,
            p2pStreams=True
            )
    return daemon

@pytest.fixture()
def ipfsdaemon2(event_loop, tmpdir):
    dir = tmpdir.mkdir('ipfsdaemon2')
    daemon = asyncipfsd.AsyncIPFSDaemon(dir,
            apiport=apiport+10,
            gatewayport=gwport+10,
            swarmport=swarmport+10,
            loop=event_loop,
            pubsubEnable=False,
            p2pStreams=True
            )
    return daemon

@pytest.fixture()
def iclient(event_loop):
    c = aioipfs.AsyncIPFS(port=apiport, loop=event_loop)
    return c

@pytest.fixture()
def ipfsop(iclient):
    return IPFSOperator(iclient, debug=True)

class EchoProtocol(asyncio.Protocol):
    def __init__(self, exitF):
        super().__init__()
        self.exitF = exitF
        self.msgCount = 0

    def connection_made(self, transport):
        print('Connection made')
        self.transport = transport
        self.transport.write('Hello'.encode())

    def data_received(self, data):
        self.msgCount += 1
        if self.msgCount < 5:
            self.transport.write(data)
        else:
            self.transport.close()

    def eof_received(self):
        pass

    def connection_lost(self, exc):
        self.exitF.set_result(True)

class TestIPFSD:
    @pytest.mark.asyncio
    async def test_basic(self, event_loop, ipfsdaemon, iclient, ipfsop):
        async def tests(op):
            id = await op.client.core.id()

            slashList = await op.filesList('/')
            async for r in op.client.add_json({'a': 123}):
                assert await op.filesLink(r, '/') == True

            await op.client.close()
            ipfsdaemon.stop()
            await asyncio.sleep(2)

        def cbstarted(f):
            event_loop.create_task(tests(ipfsop))

        started = await ipfsdaemon.start()
        ipfsdaemon.proto.startedFuture.add_done_callback(cbstarted)
        assert started == True
        await asyncio.wait([ipfsdaemon.exitFuture])

    @pytest.mark.asyncio
    async def test_tunnel(self, event_loop, ipfsdaemon, ipfsdaemon2,
            iclient, ipfsop):
        # Connect two nodes, create a P2P listener on the first one and make
        # the second node do the stream dial.

        iclient2 = aioipfs.AsyncIPFS(port=ipfsdaemon2.apiport, loop=event_loop)
        protoF = asyncio.Future()
        proto = EchoProtocol(protoF)
        port = 10010
        protoName = 'test'
        p2pL = P2PListener(iclient, protoName, ('127.0.0.1', port),
                lambda: proto)

        async def tcpEchoClient(portdest, message, loop):
            #reader, writer = await asyncio.open_connection('127.0.0.1', port)
            p = EchoProtocol(asyncio.Future())
            coro = loop.create_connection(
                    lambda: p,
                    '127.0.0.1', portdest)
            await coro

            await asyncio.sleep(2)

            encoded = message.encode()
            writer.write(encoded)
            data = await reader.read(100)
            print('Received: %r' % data.decode())
            assert data == encoded
            writer.close()

        async def createTunnel(op):
            id1 = await iclient.id()
            id2 = await iclient2.id()

            # Swarm connect the two nodes
            await iclient2.swarm.connect(
                '/ip4/127.0.0.1/tcp/{0}/ipfs/{1}'.format(
                    ipfsdaemon.swarmport, id1['ID']))

            peers2 = await iclient2.swarm.peers()
            listener = await p2pL.open()
            await iclient.p2p.listener_ls(headers=True)

            assert listener['Protocol'] == '/p2p/{}'.format(protoName)
            assert listener['Address'] == '/ip4/127.0.0.1/tcp/{}'.format(port)

            listenPort = port+1
            r = await iclient2.p2p.stream_dial(id1['ID'], 'test',
                    address='/ip4/127.0.0.1/tcp/{}'.format(listenPort))
            addr = r['Address']

            await tcpEchoClient(listenPort, 'Hello', event_loop)
            streams = await iclient2.p2p.stream_ls(headers=True)

            assert len(streams['Streams']) > 0
            print('streams', streams)
            stream = streams['Streams'].pop()

            assert stream['Protocol'] == '/p2p/{}'.format(protoName)
            assert stream['LocalPeer'] == id2['ID']
            assert stream['LocalAddress'] == '/ip4/127.0.0.1/tcp/{}'.format(
                    listenPort)
            assert stream['RemotePeer'] == id2['ID']

        def cb2started(f):
            event_loop.create_task(createTunnel(ipfsop))

        def cbstarted(f):
            ipfsdaemon2.startedFuture.add_done_callback(cb2started)

        ipfsdaemon.noBootstrap = True
        ipfsdaemon2.noBootstrap = True

        started = await ipfsdaemon.start()
        started = await ipfsdaemon2.start()

        ipfsdaemon.startedFuture.add_done_callback(cbstarted)

        await asyncio.wait([protoF])
        await p2pL.close()

        ipfsdaemon.stop()
        ipfsdaemon2.stop()

        await asyncio.wait([ipfsdaemon.exitFuture, ipfsdaemon2.exitFuture])

        await iclient.close()
        await iclient2.close()

@pytest.fixture()
def configD(tmpdir):
    return ipfsdconfig.getDefault()

class TestConfig:
    def test_default(self, configD):
        cfgStr = str(configD)
        assert 'API' in configD.c
        assert 'Bootstrap' in configD.c




import os.path

from PyQt5.QtCore import (pyqtSignal, QObject)

from galacteek.ipfs.wrappers import *

class UserProfile(QObject):
    """ User profile object """

    P_FILES = 'files'

    def __init__(self, ctx, name, rootDir):
        super(UserProfile, self).__init__()

        self._ctx = ctx
        self._name = name
        self._rootDir = rootDir

        self._filesModel = None

    @property
    def root(self):
        return self._rootDir # profile's root dir in the IPFS repository

    @property
    def ctx(self):
        return self._ctx # IPFS context

    @property
    def name(self):
        return self._name

    @property
    def filesModel(self):
        return self._filesModel

    @property
    def tree(self):
        # Profile's filesystem tree list
        return [
            self.pathFiles,
            self.pathHome,
            self.pathMedia,
            self.pathEtc,
            self.pathData,
            self.pathPlaylists,
            self.pathPictures,
            self.pathVideos,
            self.pathDocuments,
            self.pathMusic,
            self.pathCode,
            self.pathWebsites
        ]

    @property
    def pathFiles(self):
        return os.path.join(self.root, self.P_FILES)

    @property
    def pathHome(self):
        return os.path.join(self.pathFiles, 'home')

    @property
    def pathDocuments(self):
        return os.path.join(self.pathFiles, 'documents')

    @property
    def pathMedia(self):
        return os.path.join(self.pathFiles, 'multimedia')

    @property
    def pathPictures(self):
        return os.path.join(self.pathMedia, 'pictures')

    @property
    def pathVideos(self):
        return os.path.join(self.pathMedia, 'videos')

    @property
    def pathMusic(self):
        return os.path.join(self.pathMedia, 'music')

    @property
    def pathCode(self):
        return os.path.join(self.pathFiles, 'code')

    @property
    def pathWebsites(self):
        return os.path.join(self.pathFiles, 'websites')

    @property
    def pathEtc(self):
        return os.path.join(self.root, 'etc')

    @property
    def pathData(self):
        return os.path.join(self.root, 'data')

    @property
    def pathPlaylists(self):
        return os.path.join(self.pathData, 'playlists')

    def setFilesModel(self, model):
        self._filesModel = model

    @ipfsOp
    async def init(self, ipfsop):
        """
        Initialize the profile's filesystem
        """

        await ipfsop.filesMkdir(self.root)

        for directory in self.tree:
            await ipfsop.filesMkdir(directory)

        # If an old profile-less files entry is found at the root,
        # move it to the home if this is the default profile
        exFiles = await ipfsop.filesLookup(GFILES_ROOT_PATH, 'myfiles')
        if exFiles and self.name == 'default':
            r = await ipfsop.filesMove(self.pathHome,
                    self.pathHome + '.old')
            r = await ipfsop.filesMove(
                    os.path.join(GFILES_ROOT_PATH, 'myfiles'),
                    self.pathHome)




import time

from PyQt5.QtWidgets import (QWidget, QApplication,
        QDialog, QLabel, QTextEdit, QPushButton, QMessageBox)

from PyQt5.QtCore import QUrl, Qt, pyqtSlot, QCoreApplication
from PyQt5.QtGui import QClipboard

from galacteek.core.ipfsmarks import *
from galacteek.ipfs import cidhelpers

from . import ui_addkeydialog, ui_addhashmarkdialog
from . import ui_addfeeddialog
from . import ui_ipfscidinputdialog, ui_ipfsmultiplecidinputdialog
from . import ui_donatedialog

import mimetypes

def boldLabelStyle():
    return 'QLabel { font-weight: bold; }'

class AddHashmarkDialog(QDialog):
    def __init__(self, marks, resource, title, description, stats, parent=None):
        super().__init__(parent)

        self.ipfsResource = resource
        self.marks = marks
        self.stats = stats if stats else {}

        self.ui = ui_addhashmarkdialog.Ui_AddHashmarkDialog()
        self.ui.setupUi(self)
        self.ui.resourceLabel.setText(self.ipfsResource)
        self.ui.resourceLabel.setStyleSheet(boldLabelStyle())
        self.ui.newCategory.textChanged.connect(self.onNewCatChanged)
        self.ui.title.setText(title)

        if isinstance(description, str):
            self.ui.description.insertPlainText(description)

        for cat in self.marks.getCategories():
            self.ui.category.addItem(cat)

    def onNewCatChanged(self, text):
        if len(text) > 0:
            self.ui.category.setEnabled(False)
        else:
            self.ui.category.setEnabled(True)

    def accept(self):
        share = self.ui.share.isChecked()
        newCat = self.ui.newCategory.text()
        description = self.ui.description.toPlainText()

        if len(newCat) > 0:
            category = newCat
        else:
            category = self.ui.category.currentText()

        mark = IPFSHashMark.make(self.ipfsResource,
            title=self.ui.title.text(),
            share=share,
            comment=self.ui.comment.text(),
            description=description,
            tags=self.ui.tags.text().split(),
            datasize=self.stats.get('DataSize', None),
            cumulativesize=self.stats.get('CumulativeSize', None),
            numlinks=self.stats.get('NumLinks', None)
        )

        self.marks.insertMark(mark, category)
        self.done(0)

class AddFeedDialog(QDialog):
    def __init__(self, marks, resource, parent=None):
        super().__init__(parent)

        self.marks = marks
        self.ipfsResource = resource

        self.ui = ui_addfeeddialog.Ui_AddFeedDialog()
        self.ui.setupUi(self)
        self.ui.resourceLabel.setText(self.ipfsResource)
        self.ui.resourceLabel.setStyleSheet(boldLabelStyle())

    def accept(self):
        share = self.ui.share.isChecked()

        self.marks.follow(self.ipfsResource, self.ui.feedName.text(),
            resolveevery=self.ui.resolve.value(),
            share=share)
        self.done(0)

class IPFSCIDInputDialog(QDialog):
    """ Dialog for IPFS CID input and validation """

    def __init__(self, parent=None):
        super().__init__(parent)

        self.validCid = False
        self.ui = ui_ipfscidinputdialog.Ui_CIDInputDialog()
        self.ui.setupUi(self)
        self.ui.validity.setStyleSheet(boldLabelStyle())
        self.ui.cid.textChanged.connect(self.onCIDChanged)
        self.ui.clearButton.clicked.connect(self.onClearCID)

    def onClearCID(self):
        self.clearCID()

    def clearCID(self):
        """ Clears the CID input text """
        self.ui.cid.clear()

    def onCIDChanged(self, text):
        """
        When the CID input text changes we verify if it's a valid CID and
        update the validity label
        """
        if cidhelpers.cidValid(text):
            cid = self.getCID()
            if cid:
                self.ui.validity.setText(
                    'Valid CID version {}'.format(cid.version))
                self.validCid = True
            else:
                # Unlikely
                self.ui.validity.setText('Unknown CID type')
                self.validCid = False
        else:
            self.ui.validity.setText('Invalid CID')
            self.validCid = False

    def getHash(self):
        """ Returns the hash corresponding to the input, if valid """
        cid = self.getCID()
        if cid:
            return str(cid)

    def getCID(self):
        """ Returns the CID object corresponding to the input """
        try:
            return cidhelpers.getCID(self.ui.cid.text())
        except:
            return None

    def accept(self):
        if self.validCid is True:
            self.done(1)

class IPFSMultipleCIDInputDialog(QDialog):
    """ Dialog for multiple IPFS CID input """

    def __init__(self, parent=None):
        super().__init__(parent)

        self.validCid = False
        self.ui = ui_ipfsmultiplecidinputdialog.Ui_MultipleCIDInputDialog()
        self.ui.setupUi(self)
        self.ui.validity.setStyleSheet(boldLabelStyle())
        self.ui.cid.textChanged.connect(self.onCIDChanged)
        self.ui.addCidButton.clicked.connect(self.onAddCID)

    def onAddCID(self):
        if self.validCid:
            self.ui.cidList.addItem(self.ui.cid.text())
            self.ui.cid.clear()

    def onCIDChanged(self, text):
        if cidhelpers.cidValid(text):
            self.ui.validity.setText('Valid CID')
            self.validCid = True
        else:
            self.ui.validity.setText('Invalid CID')
            self.validCid = False

    def getCIDs(self):
        cids = []
        for idx in range(0, self.ui.cidList.count()):
            item = self.ui.cidList.item(idx)
            cids.append(item.text())
        return cids

    def accept(self):
        self.done(1)

class DonateDialog(QDialog):
    def __init__(self, moneroAddr, bcAddr, parent=None):
        super().__init__(parent)

        self.ui = ui_donatedialog.Ui_DonateDialog()
        self.ui.setupUi(self)
        self.ui.moneroClip.clicked.connect(self.onMoneroClip)
        self.ui.okButton.clicked.connect(self.close)
        self.ui.bitcoinClip.clicked.connect(self.onBcClip)

        self._moneroAddr = moneroAddr
        self._bcAddr = bcAddr
        self.ui.moneroAddress.setText('<b>{0}</b>'.format(self._moneroAddr))
        self.ui.bitcoinAddress.setText('<b>{0}</b>'.format(self._bcAddr))
        self.setWindowTitle('Make a donation')

    def onMoneroClip(self):
        self.toClip(self._moneroAddr)

    def onBcClip(self):
        self.toClip(self._bcAddr)

    def toClip(self, data):
        app = QCoreApplication.instance()
        app.clipboard().setText(data, QClipboard.Selection)
        app.clipboard().setText(data, QClipboard.Clipboard)

    def close(self):
        self.done(1)

    def accept(self):
        self.done(1)



import json
import collections

class Encoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, PubsubMessage):
            return obj.data
        if isinstance(obj, MarksBroadcastMessage):
            return obj.data
        return json.JSONEncoder.default(self, obj)

class PubsubMessage(collections.UserDict):
    def __str__(self):
        return json.dumps(self.data, indent=4, cls=Encoder)

class MarksBroadcastMessage(PubsubMessage):
    TYPE = 'marksbroadcast'

    @staticmethod
    def make(marksdict):
        msg = MarksBroadcastMessage({
            'msgtype': MarksBroadcastMessage.TYPE,
            'marks': marksdict
        })
        return msg




import sys
import os.path
import json
import tempfile

from async_generator import async_generator, yield_

import aioipfs
import asyncio

GFILES_ROOT_PATH = '/galacteek/'
GFILES_MYFILES_PATH = os.path.join(GFILES_ROOT_PATH, 'myfiles')
GFILES_WEBSITES_PATH = os.path.join(GFILES_ROOT_PATH, 'websites')

def joinIpfs(path):
    return os.path.join('/ipfs/', path)

def stripIpfs(path):
    return path.strip('/ipfs/')

def joinIpns(path):
    return os.path.join('/ipns/', path)

class OperatorException(Exception):
    pass

class IPFSLogWatcher(object):
    def __init__(self, operator):
        self.op = operator

    async def analyze(self):
        import pprint
        async for msg in self.op.client.log.tail():
            event = msg.get('event', None)
            time = msg.get('time', None)

            if not event:
                continue

            if event == 'handleAddProvider':
                # Handle add provider
                if self.op.ctx:
                    self.op.ctx.logAddProvider.emit(msg)

class IPFSOperator(object):
    def __init__(self, client, ctx=None, debug=False):
        self.client = client
        self.debugInfo = debug
        self.ctx = ctx

        self.filesChroot = None

    async def sleep(self, t=0):
        await asyncio.sleep(t)

    @property
    def logwatch(self):
        return IPFSLogWatcher(self)

    def debug(self, msg):
        if self.debugInfo:
            print(msg, file=sys.stderr)

    async def filesDelete(self, path, name, recursive=False):
        try:
            ret = await self.client.files.rm(os.path.join(path, name),
                    recursive=recursive)
        except aioipfs.APIException as exc:
            self.debug('Exception on removing {0} in {1}'.format(
                name, path))
            return False
        await self.client.files.flush(path)
        return True

    async def filesLookup(self, path, name):
        """
        Looks for a file (files API) with a given name in this path

        :param str path: the path to search
        :param str name: the entry name to look for
        :return: IPFS entry
        """
        listing = await self.filesList(path)
        for entry in listing:
            if entry['Name'] == name:
                return entry

    async def filesLookupHash(self, path, hash):
        """
        Searches for a file (files API) with a given hash in this path

        :param str path: the path to search
        :param str hash: the multihash to look for
        :return: IPFS entry
        """
        listing = await self.filesList(path)
        for entry in listing:
            if entry['Hash'] == hash:
                return entry

    async def filesRm(self, path, recursive=False):
        try:
            await self.client.files.rm(path, recursive=recursive)
        except aioipfs.APIException as exc:
            self.debug(exc.message)
            return None

    async def filesMkdir(self, path, parents=True):
        try:
            await self.client.files.mkdir(path, parents=parents)
        except aioipfs.APIException as exc:
            self.debug(exc.message)
            return None

    async def filesMove(self, source, dest):
        try:
            return await self.client.files.mv(source, dest)
        except aioipfs.APIException as exc:
            self.debug(exc.message)
            return None

    async def filesList(self, path):
        try:
            listing = await self.client.files.ls(path, long=True)
        except aioipfs.APIException as exc:
            self.debug(exc.message)
            return None

        if not 'Entries' in listing or listing['Entries'] is None:
            return []

        return listing['Entries']

    async def filesStat(self, path):
        try:
            return await self.client.files.stat(path)
        except aioipfs.APIException as exc:
            self.debug(exc.message)
            return None

    async def chroot(self, path):
        self.filesChroot = path

    async def vMkdir(self, path):
        if self.filesChroot:
            return await self.filesMkdir(os.path.join(self.filesChroot,
                path))

    async def vFilesList(self, path):
        if self.filesChroot:
            return await self.filesList(os.path.join(self.filesChroot,
                path))
        else:
            raise OperatorException('No chroot provided')

    async def filesLink(self, entry, dest, flush=True, name=None):
        """ Given an entry (as returned by /files/ls), make a link
            in dest """

        try:
            resp = await self.client.files.cp(joinIpfs(entry['Hash']),
                os.path.join(dest, name if name else entry['Name']))
        except aioipfs.APIException as exc:
            self.debug('Exception on copying entry {0} to {1}'.format(
                entry, dest))
            return False

        if flush:
            await self.client.files.flush(dest)
        return True

    async def filesLinkFp(self, entry, dest):
        try:
            resp = await self.client.files.cp(joinIpfs(entry['Hash']),
                    dest)
        except aioipfs.APIException as exc:
            self.debug('Exception on copying entry {0} to {1}'.format(
                entry, dest))
            return False

        return True

    async def peersList(self):
        try:
            peers = await self.client.swarm.peers()
        except aioipfs.APIException as exc:
            return None

        if peers and 'Peers' in peers:
            return peers.get('Peers', None)

    async def nodeId(self):
        info = await self.client.core.id()
        return info.get('ID', 'Unknown')

    async def keysNames(self):
        keys = await self.keys()
        return [key['Name'] for key in keys]

    async def keys(self):
        kList = await self.client.key.list(long=True)
        return kList.get('Keys', [])

    async def keysRemove(self, name):
        try:
            await self.client.key.rm(name)
        except aioipfs.APIException as exc:
            self.debug('Exception on removing key {0}: {1}'.format(
                name, exc.message))
            return False
        return True

    async def publish(self, path, key='self'):
        try:
            return await self.client.name.publish(path, key=key)
        except aioipfs.APIException as exc:
            return None

    async def resolve(self, path, timeout=20):
        try:
            resolved = await asyncio.wait_for(self.client.name.resolve(path),
                    timeout)
        except asyncio.TimeoutError as timeExc:
            return None
        except aioipfs.APIException as e:
            self.debug(e.message)
            return None
        else:
            return resolved

    async def purge(self, hashRef, rungc=False):
        """ Unpins an object and optionally runs the garbage collector """
        try:
            await self.client.pin.rm(hashRef, recursive=True)
            if rungc:
                await self.client.repo.gc()
            return True
        except aioipfs.APIException as e:
            return False

    async def isPinned(self, hashRef):
        """
        Returns True if IPFS object references by hashRef is pinned,
        False otherwise
        """
        try:
            result = await self.client.pin.ls(multihash=hashRef)
            keys = result.get('Keys', {})
            return key in keys
        except aioipfs.APIException as e:
            return False

    async def pinned(self, type='all'):
        """
        Returns all pinned keys of a given type
        """
        try:
            result = await self.client.pin.ls(pintype=type)
            return result.get('Keys', {})
        except aioipfs.APIException as e:
            return None

    @async_generator
    async def list(self, path, resolve_type=True):
        """
        Lists objects in a given path and yields them
        """
        try:
            listing = await self.client.ls(path, headers=True,
                    resolve_type=resolve_type)
            objects = listing.get('Objects', [])

            for obj in objects:
                await self.sleep()
                await yield_(obj)
        except:
            pass

    async def objStat(self, path, timeout=30):
        try:
            stat = await asyncio.wait_for(self.client.object.stat(path),
                    timeout)
        except Exception as e:
            return None
        else:
            return stat

    async def objStatCtxUpdate(self, path):
        try:
            self.ctx.objectStats[path] = await self.objStat(path)
            return self.ctx.objectStats[path]
        except aioipfs.APIException as e:
            self.ctx.objectStats[path] = None

    def objStatCtxGet(self, path):
        return self.ctx.objectStats.get(path, None)

    async def addPath(self, path, **kw):
        """
        Recursively adds files from path, and returns the top-level entry (the
        root directory), optionally wrapping it with a directory object

        :param str path: the path to the directory/file to import
        :param bool wrap: add a wrapping directory
        :return: the IPFS top-level entry
        :rtype: dict
        """
        added = None
        entryCb = kw.pop('callback', None)
        async for entry in self.client.add(path, quiet=True,
                recursive=kw.pop('recursive', True),
                wrap_with_directory=kw.pop('wrap', False)):
            await self.sleep()
            added = entry
            if asyncio.iscoroutinefunction(entryCb):
                await entryCb(entry)
        return added

    async def closestPeers(self):
        peers = []
        try:
            info = await self.client.core.id()
            queryR = await self.client.dht.query(info['ID'])

            responses = queryR.get('Responses', None)
            if not responses:
                return None
            for resp in responses:
                peers.append(resp['ID'])
            return peers
        except aioipfs.APIException as e:
            return None

    async def jsonLoad(self, path):
        try:
            data = await self.client.cat(path)
            return json.loads(data)
        except aioipfs.APIException as exc:
            self.debug(exc.message)
            return None

    async def dagPut(self, data):
        """
        Create a new DAG object from data and returns the root hash of the DAG
        """
        dagFile = tempfile.NamedTemporaryFile()
        if not dagFile:
            return None
        try:
            dagFile.write(json.dumps(data).encode())
        except Exception as e:
            self.debug('Cannot convert DAG object: {}'.format(str(e)))
            return None

        dagFile.seek(0, 0)
        try:
            output = await self.client.dag.put(dagFile.name)
            if 'Cid' in output:
                return output['Cid'].get('/', None)
        except aioipfs.APIException as exc:
            self.debug(exc.message)
            return None

    async def dagGet(self, multihash):
        """
        Gets the DAG object referenced by multihash and returns a JSON object
        """
        try:
            output = await self.client.dag.get(multihash)
            return json.loads(output)
        except aioipfs.APIException as exc:
            self.debug(exc.message)
            return None




import sys
import json

import aioipfs
import asyncio
import socket

class P2PProtocol(asyncio.Protocol):
    def __init__(self):
        super(P2PProtocol, self).__init__()

        self.exitFuture = asyncio.Future()
        self.eofReceived = False

    def connection_made(self, transport):
        self.transport = transport

    def data_received(self, data):
        self.transport.write(data)

    def eof_received(self):
        self.eofReceived = True

    def connection_lost(self, exc):
        self.exitFuture.set_result(True)

class P2PListener(object):
    """ IPFS P2P listener class """

    def __init__(self, client, protocol, address, factory):
        self.client = client
        self._protocol = protocol
        self._address = address
        self._factory = factory
        self._server = None

        self.loop = asyncio.get_event_loop()

    @property
    def server(self):
        return self._server

    @property
    def protocol(self):
        return self._protocol

    @property
    def address(self):
        return self._address

    @property
    def protocolFactory(self):
        return self._factory

    async def open(self):
        listenAddress = '/ip4/{addr}/tcp/{port}'.format(
            addr=self.address[0],
            port=self.address[1]
        )

        try:
            addr = await self.client.p2p.listener_open(
                self.protocol, listenAddress)
        except aioipfs.APIException as exc:
            # P2P not enabled or some other reason
            return None
        else:
            # Now that the listener is registered, create the server socket
            # and return the listener's address
            self._server = await self.loop.create_server(self.protocolFactory,
                self.address[0], self.address[1])
            return addr

    async def close(self):
        ret = await self.client.p2p.listener_close(self.protocol)

        if self._server:
            self._server.close()
            await self._server.wait_closed()




import sys
import json
import os, os.path
import multiprocessing
import time
import asyncio
import re
import collections
import pkg_resources
import jinja2, jinja2.exceptions

from quamash import QEventLoop

from PyQt5.QtWidgets import QApplication, QSystemTrayIcon, QMenu
from PyQt5.QtGui import QPixmap, QIcon, QClipboard
from PyQt5.QtCore import (QCoreApplication, QUrl, QStandardPaths,
        QSettings, QTranslator, QFile, pyqtSignal, QObject,
        QTemporaryDir, QDateTime, QMessageLogger, QMimeDatabase)

from galacteek.ipfs import pinning, ipfsd, asyncipfsd, cidhelpers
from galacteek.ipfs.ipfsops import *
from galacteek.ipfs.wrappers import *
from galacteek.ipfs.pubsub import *
from galacteek.ipfs.feeds import FeedFollower

from galacteek.ui import mainui, galacteek_rc, downloads, browser, hashmarks
from galacteek.ui.helpers import *

from galacteek.appsettings import *
from galacteek.core.ipfsmarks import IPFSMarks
from galacteek.core.profile import UserProfile

from yarl import URL

import aioipfs

GALACTEEK_NAME = 'galacteek'

# Application's i18n messages

# IPFS daemon messages

def iIpfsDaemonStarted():
    return QCoreApplication.translate('Galacteek', 'IPFS daemon started')

def iIpfsDaemonGwStarted():
    return QCoreApplication.translate('Galacteek',
        "IPFS daemon's gateway started")

def iIpfsDaemonReady():
    return QCoreApplication.translate('Galacteek', 'IPFS daemon is ready')

def iIpfsDaemonProblem():
    return QCoreApplication.translate('Galacteek',
        'Problem starting IPFS daemon')

def iIpfsDaemonInitProblem():
    return QCoreApplication.translate('Galacteek',
        'Problem initializing the IPFS daemon (check the ports configuration)')

def iIpfsDaemonWaiting(count):
    return QCoreApplication.translate('Galacteek',
        'IPFS daemon: waiting for connection (try {0})'.format(count))

class IPFSConnectionParams(object):
    def __init__(self, host, apiport, gwport):
        self._host = host
        self._apiPort = apiport
        self._gatewayPort = gwport

        self._gatewayUrl = URL.build(host=self.host,
            port=self.gatewayPort, scheme='http', path='')

    @property
    def host(self):
        return self._host

    @property
    def apiPort(self):
        return self._apiPort

    @property
    def gatewayPort(self):
        return self._gatewayPort

    @property
    def gatewayUrl(self):
        return self._gatewayUrl

class IPFSContext(QObject):
    # signals
    ipfsRepositoryReady = pyqtSignal()

    # profiles
    profilesAvailable = pyqtSignal(list)
    profileChanged = pyqtSignal(str)

    # log events
    logAddProvider = pyqtSignal(dict)

    # pubsub
    pubsubMessageRx = pyqtSignal()
    pubsubMessageTx = pyqtSignal()

    pubsubMarksReceived = pyqtSignal(int)

    # pinning signals
    pinQueueSizeChanged = pyqtSignal(int)
    pinItemStatusChanged = pyqtSignal(str, dict)
    pinItemsCount = pyqtSignal(int)
    pinNewItem = pyqtSignal(str)
    pinFinished = pyqtSignal(str)

    def __init__(self):
        super().__init__()

        self.objectStats = {}
        self.profiles = {}
        self.currentProfile = None
        self.ipfsClient = None

    @property
    def client(self):
        return self.ipfsClient

    def setIpfsClient(self, client):
        self.ipfsClient = client

    @ipfsOp
    async def profilesInit(self, ipfsop):
        hasGalacteek = await ipfsop.filesLookup('/', GALACTEEK_NAME)
        if not hasGalacteek:
            await ipfsop.client.files.mkdir(GFILES_ROOT_PATH, parents=True)

        rootList = await ipfsop.filesList(GFILES_ROOT_PATH)

        # Scans existing profiles
        for entry in rootList:
            name = entry['Name']
            if entry['Type'] == 1 and name.startswith('profile.'):
                ma = re.search('profile\.([a-zA-Z\.\_\-]*)$', name)
                if ma:
                    profileName = ma.group(1).rstrip()
                    profile = await self.profileNew(profileName)

        defaultProfile = 'default'

        # Create default profile if not found
        if defaultProfile not in self.profiles:
            profile = await self.profileNew(defaultProfile)

        self.profileEmitAvail()
        self.profileChange(defaultProfile)

        self.ipfsRepositoryReady.emit()

    def profileGet(self, name):
        return self.profiles.get(name, None)

    def profileEmitAvail(self):
        self.profilesAvailable.emit(list(self.profiles.keys()))

    async def profileNew(self, pName):
        profile = UserProfile(self, pName,
                os.path.join(GFILES_ROOT_PATH, 'profile.{}'.format(pName)))
        try:
            await profile.init()
        except Exception as e:
            return None
        self.profiles[pName] = profile
        self.profileEmitAvail()
        return profile

    def profileChange(self, pName):
        if pName in self.profiles:
            self.currentProfile = self.profiles[pName]
            self.profileChanged.emit(pName)
            return True
        else:
            return False

class GalacteekApplication(QApplication):
    clipboardHasIpfs = pyqtSignal(bool, str, str)
    manualAvailable = pyqtSignal(str, dict)

    def __init__(self, debug=False, profile='main'):
        QApplication.__init__(self, sys.argv)

        QCoreApplication.setApplicationName(GALACTEEK_NAME)

        self._appProfile = profile
        self._loop = None
        self._ipfsClient = None
        self._ipfsd = None

        self.translator = None
        self.logger = QMessageLogger()

        self.mainWindow = None
        self.gWindows = []

        self.pubsubListeners = []
        self.pinnerTask = None

        self.debugEnabled = debug

        self.ipfsCtx = IPFSContext()

        self.setupPaths()
        self.setupClipboard()
        self.setupSchemeHandlers()

        self.initSettings()
        self.setupTranslator()
        self.initSystemTray()
        self.initMisc()
        self.createMainWindow()

        self.setStyle()
        self.clipboardInit()

    @property
    def appProfile(self):
        return self._appProfile

    @property
    def ipfsd(self):
        return self._ipfsd

    @property
    def loop(self):
        return self._loop

    @property
    def ipfsClient(self):
        return self._ipfsClient

    @property
    def gatewayAuthority(self):
        params = self.getIpfsConnectionParams()
        return '{0}:{1}'.format(params.host, params.gatewayPort)

    @property
    def gatewayUrl(self):
        params = self.getIpfsConnectionParams()
        return params.gatewayUrl

    @property
    def ipfsDataLocation(self):
        return self._ipfsDataLocation

    def setStyle(self):
        qssPath = ":/share/static/qss/galacteek.qss"
        qssFile = QFile(qssPath)

        try:
            qssFile.open(QFile.ReadOnly)
            styleSheetBa = qssFile.readAll()
            styleSheetStr = styleSheetBa.data().decode('utf-8')
            self.setStyleSheet(styleSheetStr)
        except:
            # that would probably occur if the QSS is not in the resources file..
            # set some default stylesheet here?
            pass

    def debug(self, msg):
        if self.debugEnabled:
            self.logger.debug(msg)

    def initSystemTray(self):
        self.systemTray = QSystemTrayIcon(self)
        self.systemTray.setIcon(getIcon('ipfs-logo-128-ice-text.png'))
        self.systemTray.show()
        self.systemTray.activated.connect(self.onSystemTrayIconClicked)
        self.systemTray.setToolTip('Galacteek')

        systemTrayMenu = QMenu(None)

        action = systemTrayMenu.addAction('Quit')
        action.setIcon(QIcon.fromTheme('application-exit'))
        action.triggered.connect(self.onExit)

        self.systemTray.setContextMenu(systemTrayMenu)

    def initMisc(self):
        self.jinjaEnv = jinja2.Environment(
            loader=jinja2.PackageLoader('galacteek', 'templates'))

        self.manuals = ManualsImporter(self)
        self.mimeDb = QMimeDatabase()

        self.downloadsManager = downloads.DownloadsManager(self)
        self.marksLocal = IPFSMarks(self.localMarksFileLocation)
        self.importDefaultHashmarks(self.marksLocal)

        self.marksLocal.addCategory('general')
        self.marksNetwork = IPFSMarks(self.networkMarksFileLocation)

        self.tempDir = QTemporaryDir()
        if not self.tempDir.isValid():
            pass

    def importDefaultHashmarks(self, marksLocal):
        pkg = 'galacteek.hashmarks.default'
        try:
            listing = pkg_resources.resource_listdir(pkg, '')
            for fn in listing:
                if fn.endswith('.json'):
                    path = pkg_resources.resource_filename(pkg, fn)
                    self.debug('Importing hashmark file: {}'.format(path))
                    marks = IPFSMarks(path)
                    marksLocal.merge(marks)
        except Exception as e:
            self.debug(str(e))

    def setupTranslator(self):
        if self.translator:
            QApplication.removeTranslator(self.translator)

        self.translator = QTranslator()
        QApplication.installTranslator(self.translator)
        lang = self.settingsMgr.getSetting(CFG_SECTION_UI, CFG_KEY_LANG)
        self.translator.load(':/share/translations/galacteek_{0}.qm'.format(
            lang))

    def createMainWindow(self, show=True):
        self.mainWindow = mainui.MainWindow(self)
        if show is True:
            self.mainWindow.show()

    def onSystemTrayIconClicked(self, reason):
        if reason == QSystemTrayIcon.Unknown:
            pass
        elif reason == QSystemTrayIcon.Context:
            pass
        elif reason == QSystemTrayIcon.DoubleClick:
            self.mainWindow.show()
        else:
            pass

    def systemTrayMessage(self, title, message, timeout=2000,
            messageIcon=QSystemTrayIcon.Information):
        self.systemTray.showMessage(title, message, messageIcon, timeout)

    def setupRepository(self):
        self.task(self.ipfsCtx.profilesInit)
        self.manuals.importManualMain()

    def ipfsTask(self, fn, *args, **kw):
        """ Schedule an async IPFS task """
        return self.loop.create_task(fn(self.ipfsClient,
            *args, **kw))

    def ipfsTaskOp(self, fn, *args, **kw):
        """ Schedule an async IPFS task using an IPFSOperator instance """
        client = self.ipfsClient
        if client:
            return self.loop.create_task(fn(
                self.getIpfsOperator(), *args, **kw))

    def setIpfsClient(self, client):
        self._ipfsClient = client

    def getIpfsClient(self):
        return self._ipfsClient

    def getIpfsOperator(self):
        return IPFSOperator(self.ipfsClient, ctx=self.ipfsCtx,
            debug=self.debugEnabled)

    def getIpfsConnectionParams(self):
        mgr = self.settingsMgr
        section = CFG_SECTION_IPFSD
        if mgr.isTrue(section, CFG_KEY_ENABLED):
            return IPFSConnectionParams('localhost',
                    mgr.getSetting(section, CFG_KEY_APIPORT),
                    mgr.getSetting(section, CFG_KEY_HTTPGWPORT)
            )
        else:
            section = CFG_SECTION_IPFSCONN1
            return IPFSConnectionParams(
                    mgr.getSetting(section, CFG_KEY_HOST),
                    mgr.getSetting(section, CFG_KEY_APIPORT),
                    mgr.getSetting(section, CFG_KEY_HTTPGWPORT)
            )

    def updateIpfsClient(self):
        connParams = self.getIpfsConnectionParams()
        client = aioipfs.AsyncIPFS(host=connParams.host,
                port=connParams.apiPort, loop=self.loop)
        self.setIpfsClient(client)
        self.ipfsCtx.setIpfsClient(client)
        self.setupRepository()
        self.setupIpfsServices()

    def setupIpfsServices(self):
        pubsubEnabled = self.settingsMgr.isTrue(CFG_SECTION_IPFS,
            CFG_KEY_PUBSUB)
        if pubsubEnabled:
            # Main pubsub topic
            listenerMain = PubsubListener(self.ipfsClient,
                self.loop, self.ipfsCtx, topic='galacteek.main')
            listenerMain.start()
            self.pubsubListeners.append(listenerMain)

            # Pubsub marks exchanger
            listenerMarks = HashmarksExchanger(self.ipfsClient,
                self.loop, self.ipfsCtx, self.marksLocal,
                self.marksNetwork)
            listenerMarks.start()
            self.pubsubListeners.append(listenerMarks)

        self.feedFollower = FeedFollower(self.marksLocal)
        self.ipfsTaskOp(self.feedFollower.process)

    def stopIpfsServices(self):
        for s in self.pubsubListeners:
            s.stop()
        if self.pinnerTask is not None:
            self.pinnerTask.cancel()

    def startPinner(self):
        self.pinner = pinning.Pinner(self)
        self.pinnerTask = self.loop.create_task(self.pinner.process())

    def setupAsyncLoop(self):
        loop = QEventLoop(self)
        asyncio.set_event_loop(loop)
        self.setLoop(loop)
        return loop

    def setLoop(self, loop):
        self._loop = loop

    def getLoop(self):
        return self._loop

    def task(self, fn, *args, **kw):
        return self.loop.create_task(fn(*args, **kw))

    def setupPaths(self):
        qtDataLocation = QStandardPaths.writableLocation(QStandardPaths.DataLocation)

        if not qtDataLocation:
            raise Exception('No writable data location found')

        self._dataLocation = os.path.join(
                qtDataLocation, self._appProfile)

        self._ipfsDataLocation = os.path.join(self._dataLocation, 'ipfs')
        self.marksDataLocation = os.path.join(self._dataLocation, 'marks')
        self.localMarksFileLocation = os.path.join(self.marksDataLocation,
            'ipfsmarks.local.json')
        self.networkMarksFileLocation = os.path.join(self.marksDataLocation,
            'ipfsmarks.network.json')

        qtConfigLocation = QStandardPaths.writableLocation(
                QStandardPaths.ConfigLocation)
        self.configDirLocation = os.path.join(
                qtConfigLocation, GALACTEEK_NAME, self._appProfile)
        self.settingsFileLocation = os.path.join(
                self.configDirLocation, '{}.conf'.format(GALACTEEK_NAME))

        for dir in [ self._ipfsDataLocation,
                self.marksDataLocation,
                self.configDirLocation ]:
            if not os.path.exists(dir):
                os.makedirs(dir)

        self.defaultDownloadsLocation = QStandardPaths.writableLocation(
            QStandardPaths.DownloadLocation)

        self.debug('Data {0}, config {1}, configfile {2}'.format(
                self._dataLocation,
                self.configDirLocation,
                self.settingsFileLocation))

    def initSettings(self):
        self.settingsMgr = SettingsManager(path=self.settingsFileLocation)
        setDefaultSettings(self)
        self.settingsMgr.sync()

    def startIpfsDaemon(self, migrateRepo=False):
        if self.ipfsd: # we only support one daemon for now
            return

        pubsubEnabled = self.settingsMgr.isTrue(CFG_SECTION_IPFS,
            CFG_KEY_PUBSUB)
        corsEnabled = self.settingsMgr.isTrue(CFG_SECTION_IPFSD,
            CFG_KEY_CORS)

        sManager = self.settingsMgr
        section = CFG_SECTION_IPFSD

        # Instantiate an IPFS daemon using asyncipfsd and
        # start it in a task, monitoring the initialization process

        self._ipfsd = asyncipfsd.AsyncIPFSDaemon(self.ipfsDataLocation,
            apiport=sManager.getInt(section, CFG_KEY_APIPORT),
            swarmport=sManager.getInt(section, CFG_KEY_SWARMPORT),
            gatewayport=sManager.getInt(section, CFG_KEY_HTTPGWPORT),
            swarmLowWater=sManager.getInt(section, CFG_KEY_SWARMLOWWATER),
            swarmHighWater=sManager.getInt(section, CFG_KEY_SWARMHIGHWATER),
            storageMax=sManager.getInt(section, CFG_KEY_STORAGEMAX),
            pubsubEnable=pubsubEnabled, corsEnable=corsEnabled,
            migrateRepo=migrateRepo)

        self.task(self.startIpfsdTask, self.ipfsd)

    async def startIpfsdTask(self, ipfsd):
        res = await ipfsd.start()
        self.mainWindow.statusMessage(iIpfsDaemonStarted())

        if res:
            running = False
            self.systemTrayMessage('IPFS', iIpfsDaemonStarted(),
                timeout=1000)
            for iters in range(0, 64):
                self.mainWindow.statusMessage(iIpfsDaemonWaiting(iters))
                await asyncio.sleep(0.2)
                if ipfsd.proto.daemonReady:
                    # Good to go
                    self.systemTrayMessage('IPFS',
                        iIpfsDaemonReady(), timeout=1000)
                    running = True
                    break

            if running is True:
                self.updateIpfsClient()
            else:
                self.systemTrayMessage('IPFS', iIpfsDaemonInitProblem())
        else:
            self.systemTrayMessage('IPFS', iIpfsDaemonProblem())

    def setupClipboard(self):
        self.appClipboard = self.clipboard()
        self.clipTracker = ClipboardTracker(self.appClipboard)

    def clipboardInit(self):
        self.clipTracker.clipboardInit()

    def setClipboardText(self, text):
        self.clipTracker.setText(text)

    def getClipboardText(self):
        return self.clipTracker.getText()

    def setupSchemeHandlers(self):
        self.ipfsSchemeHandler = browser.IPFSSchemeHandler(self)

    def subUrl(self, path):
        """ Joins the gatewayUrl and path to form a new URL """
        sub = QUrl(str(self.gatewayUrl))
        sub.setPath(path)
        return sub

    def getJinjaTemplate(self, name):
        try:
            tmpl = self.jinjaEnv.get_template(name)
        except jinja2.exceptions.TemplateNotFound as e:
            return None
        else:
            return tmpl

    def onExit(self):
        self.exit()

    def exit(self):
        self.stopIpfsServices()

        if self.ipfsd:
            self.ipfsd.stop()

        self.tempDir.remove()
        self.quit()
        sys.exit(0)

class ClipboardTracker(QObject):
    """
    Tracks the system's clipboard activity and emits signals
    depending on whether or not the clipboard contains an IPFS CID or path
    """
    clipboardHasIpfs = pyqtSignal(bool, str, str)
    clipboardHistoryChanged = pyqtSignal(dict)

    def __init__(self, clipboard):
        super(ClipboardTracker, self).__init__()

        self.hasIpfs = False
        self.history = {}
        self.clipboard = clipboard
        self.clipboard.changed.connect(self.onClipboardChanged)
        self.clipboardHasIpfs.connect(self.onHasIpfs)

    def onClipboardChanged(self, mode):
        text = self.clipboard.text(mode)
        self.clipboardProcess(text, clipboardMode=mode)

    def clipboardProcess(self, text, clipboardMode=None):
        """
        Process the contents of the clipboard. If it is a valid CID/path, emit a
        signal, processed by the main window for the clipboard loader button
        """
        if not text or len(text) > 1024: # that shouldn't be worth handling
            return

        text = text.strip()
        ma = cidhelpers.ipfsRegSearchPath(text)

        if ma:
            # The clipboard contains a full IPFS path
            cid = ma.group('cid')
            if not cidhelpers.cidValid(cid):
                return
            path = ma.group('fullpath')
            self.hRecord(path)
            return self.clipboardHasIpfs.emit(True, cid, path)

        ma = cidhelpers.ipnsRegSearchPath(text)
        if ma:
            # The clipboard contains a full IPNS path
            path = ma.group('fullpath')
            self.hRecord(path)
            return self.clipboardHasIpfs.emit(True, None, path)

        ma = cidhelpers.ipfsRegSearchCid(text)
        if ma:
            # The clipboard simply contains a CID
            cid = ma.group('cid')
            if not cidhelpers.cidValid(cid):
                return

            path = joinIpfs(cid)
            self.hRecord(path)
            return self.clipboardHasIpfs.emit(True, cid, path)

        # Not a CID/path
        if clipboardMode == self.clipboardPreferredMode():
            self.clipboardHasIpfs.emit(False, None, None)

    def hLookup(self, path):
        for hTs, hItem in self.history.items():
            if hItem['path'] == path:
                return hTs, hItem

    def hRecord(self, path):
        """ Records an item in the history and emits a signal """
        now = time.time()
        itLookup = self.hLookup(path)
        self.history[now] = {
            'path': path,
            'date': QDateTime.currentDateTime()
        }
        self.clipboardHistoryChanged.emit(self.getHistory())

    def getHistory(self):
        return collections.OrderedDict(sorted(self.history.items(),
            key=lambda t: t[0]))

    def clearHistory(self):
        self.history = {}
        self.clipboardHistoryChanged.emit(self.getHistory())

    def getHistoryLatest(self):
        """ Returns latest history item """
        h = self.getHistory()
        try:
            return h.popitem(last=True)[1]
        except KeyError:
            return None

    def getCurrent(self):
        """ Returns current clipboard item """
        if self.hasIpfs:
            return self.getHistoryLatest()

    def clipboardInit(self):
        """ Used to process the clipboard's content on application's init """
        text = self.getText()
        self.clipboardProcess(text)

    def clipboardPreferredMode(self):
        return QClipboard.Selection if self.clipboard.supportsSelection() \
                else QClipboard.Clipboard

    def setText(self, text):
        """ Sets the clipboard's text content """
        self.clipboard.setText(text, self.clipboardPreferredMode())

    def getText(self):
        """ Returns clipboard's text content from the preferred source """
        return self.clipboard.text(self.clipboardPreferredMode())

    def onHasIpfs(self, valid, cid, path):
        self.hasIpfs = valid

class ManualsImporter(QObject):
    """ Imports the HTML manuals in IPFS """

    def __init__(self, app):
        super(ManualsImporter, self).__init__()

        self.app = app
        self.registry = {}

    def importManualMain(self):
        try:
            listing = pkg_resources.resource_listdir('docs.manual', '')
            for entry in listing:
                if entry.startswith('__'):
                    continue
                self.importManualLang(entry)
        except Exception as e:
            self.app.debug('Failed importing manuals {0}'.format(str(e)))

    def importManualLang(self, lang):
        try:
            docPath = pkg_resources.resource_filename('docs.manual',
                '{0}/html'.format(lang))
            self.app.task(self.importDocPath, docPath, lang)
        except Exception as e:
            self.app.debug('Failed importing manual ({0}) {1}'.format(
                lang, str(e)))

    @ipfsOp
    async def importDocPath(self, ipfsop, docPath, lang):
        docEntry = await ipfsop.addPath(docPath)
        if docEntry:
            self.registry[lang] = docEntry
            self.app.manualAvailable.emit(lang, docEntry)
